/*
 * Copyright (C) 2004 by Oliver Schulz. All rights reserved.
 * mailto:olischulz@web.de
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
 * $Log$
 */

#include <compiler.h>
#include <sys/thread.h>
#include <sys/atom.h>
#include "1wire.h"

/*
 * Note: All specifications for this Dallas' 1Wire interface implementation were took
 * from the datasheet for Dallas' DS18B20 digital temperature sensor.
 */

/* 
 * You can directly connect the data pin (DQ) of the slave devices to the port pin.
 * However, if you want to use 'parasite power' you need an additional pull-up 
 * resistor connected to the port pin and +5V. In that case the resistor values
 * depends on the number of slave devices connected to the bus. Please refer
 * to the appropriate datasheets.
 * Note: I haven't tested 'parasite power' with these functions, but I think
 * it should work either.
 */

/*
 * Use the following defines to easy change the used port pin of the microcontroller.
 * This is a little odd solution, but much smaller and faster than handling
 * variable pin addresses, believe me.
 */
#define OW_1		sbi (PORTB, 5)		/* set 1wire pin to high state */
#define OW_0		cbi (PORTB, 5)		/* set 1wire pin to low state */
#define OW_out		sbi (DDRB, 5)		/* set 1wire pin to output direction */
#define OW_in		cbi (DDRB, 5)		/* set 1wire pin to input direction */
#define OW_get  	bit_is_set (PINB, 5)	/* read current 1wire pin state */


#ifdef __GNUC__
/*
 * Here one delay function written totaly in assembler for avr-gcc and
 * clock frequency of 14.7456 MHz only!
 */
#define usDelay(us) (__Delay_5us(us/5))
static void __Delay_5us(char cDelay) __attribute__ ((naked));
static void __Delay_5us(char cDelay)
{
    u_char s;
    asm volatile ("\n" ".Lagain:" "\n\t" "nop" "\n\t"   /*      1       */
                  "nop" "\n\t"  /*      1       */
                  "ldi %0, 15" "\n"     /*      1       */
                  ".Lloop:" "\n\t" "dec %0" "\n\t"      /* 1            */
                  "breq .Lstop" "\n\t"  /* 1    2       */
                  "rjmp .Lloop" "\n"    /* 2            */
                  ".Lstop:" "\n\t" "dec %1" "\n\t"      /*      1       */
                  "breq .Lfin" "\n\t"   /*      1       2       */
                  "nop" "\n\t"  /*      1       */
                  "nop" "\n\t"  /*      1       */
                  "nop" "\n\t"  /*      1       */
                  "nop" "\n\t"  /*      1       */
                  "nop" "\n\t"  /*      1       */
                  "nop" "\n\t"  /*      1       */
                  "nop" "\n\t"  /*      1       */
                  "nop" "\n\t"  /*      1       */
                  "rjmp .Lagain" "\n"   /*      2       */
                  ".Lfin:" "\n\t" "ret" "\n\t"  /*              4       */
                  :"=&d" (s):"r"(cDelay));
}
#endif
#ifdef __IMAGECRAFT__
/* Not even tested for Imagecraft compiler... (just copied) */
#define usDelay(us) (__Delay_5us(us/5))
static void __Delay_5us(char cDelay)
{
    u_char s;
    asm ("\n" ".Lagain:" "\n\t" "nop" "\n\t"   /*      1       */
                  "nop" "\n\t"  /*      1       */
                  "ldi %s, 15" "\n"     /*      1       */
                  ".Lloop:" "\n\t" "dec %s" "\n\t"      /* 1            */
                  "breq .Lstop" "\n\t"  /* 1    2       */
                  "rjmp .Lloop" "\n"    /* 2            */
                  ".Lstop:" "\n\t" "dec %cDelay" "\n\t"      /*      1       */
                  "breq .Lfin" "\n\t"   /*      1       2       */
                  "nop" "\n\t"  /*      1       */
                  "nop" "\n\t"  /*      1       */
                  "nop" "\n\t"  /*      1       */
                  "nop" "\n\t"  /*      1       */
                  "nop" "\n\t"  /*      1       */
                  "nop" "\n\t"  /*      1       */
                  "nop" "\n\t"  /*      1       */
                  "nop" "\n\t"  /*      1       */
                  "rjmp .Lagain" "\n"   /*      2       */
                  ".Lfin:" "\n\t" "ret" "\n\t"  /*              4       */
                  );
}
#endif


/*
 * Initializes the 1wire hardware. All transactions will begin with this
 * hardware reset.
 * Returns nonzero if at least one 1wire slave is ready and connected
 * to the bus.
 */
u_char DS1WInit(void)
{
    u_char r = 0;

    NutEnterCritical();
    OW_0;                       /* Set output to 0 */
    OW_out;                     /* Set PIN to output mode */
    usDelay(480);               /* Wait 480s for bus initialize */

    OW_in;                      /* Set PIN to input mode */
    OW_1;                       /* Enable pull-up resistor */

    usDelay(70);                /* Wait 70s for presence signal */
    if (!OW_get)
        r = 1;                  /* Read device presence signal */

    usDelay(410);               /* Wait for end of initialization slot */
    NutExitCritical();

    return r;
}

/*
 * Writes one bit to the 1wire bus.
 */
void DS1WWriteBit(u_char bit)
{
    /* Assuming the bus is in idle state, i.e. PIN is in input mode  */
    NutEnterCritical();
    if (bit) {                  /* write a logical 1 */
        OW_0;                   /* Set output to 0 */
        OW_out;                 /* Set PIN to output mode */
        usDelay(10);            /* Wait 10s */
        OW_in;                  /* Release the bus */
        OW_1;                   /* Enable pull-up resistor */
        usDelay(65);            /* Wait for end of slot */
    } else {                    /* write a logical 0 */
        OW_0;                   /* Set output to 0 */
        OW_out;                 /* Set PIN to output mode */
        usDelay(65);            /* Wait 65s */
        OW_in;                  /* Release the bus */
        OW_1;                   /* Enable pull-up resistor */
        usDelay(10);            /* Wait for end of slot */
    }
    NutExitCritical();
}

/*
 * Reads and returns one bit from the bus.
 */
u_char DS1WReadBit(void)
{
    /* Assuming the bus is in idle state, i.e. PIN is in input mode  */
    u_char r = 0;
    NutEnterCritical();
    OW_0;                       /* Set output to 0 */
    OW_out;                     /* Set PIN to output mode */
    usDelay(5);                 /* Wait 5s */
    OW_in;                      /* Release the bus */
    OW_1;                       /* Enable pull-up resistor */
    usDelay(5);                 /* Wait 5s again for valid signal from slave */

    if (OW_get)                 /* Read current bus level */
        r = 1;
    usDelay(55);                /* Wait for end of slot */
    NutExitCritical();
    return r;
}

/*
 * Writes one byte (8 bits) to the 1wire bus. It simply calls
 * DS1WWriteBit eight times for each bit.
 */
void DS1WWriteByte(u_char b)
{
    u_char i;
    for (i = 8; i > 0; i--) {
        DS1WWriteBit(b & 0x01);
        b >>= 1;
    }
}

/*
 * Reads and returns one byte (8 bits) from the 1wire bus. It simply calls
 * DS1WReadBit eight times for each bit.
 */
u_char DS1WReadByte(void)
{
    u_char b = 0;
    u_char i;
    for (i = 8; i > 0; i--) {
        b >>= 1;
        if (DS1WReadBit())
            b |= 0x80;
    }
    return b;
}

/*
 * Writes a complete buffer to the bus.
 */
void DS1WWriteBuf (char* buf, size_t len)
{
    for (; len; len--, buf++)
        DS1WWriteByte (*(u_char*)buf);
}

/*
 * Reads a buffer from the bus.
 */
void DS1WReadBuf (char* buf, size_t len)
{
    for (; len; len--, buf++)
        *(u_char*)buf = DS1WReadByte();
}

/*
 * Performs a ROM search routine. It's used to find out the unique 64-bit serial 
 * codes (ROM code) of each device. On each call the ROM code of the next found
 * device is written to *ROM, so be sure to provide a valid 8 byte sized buffer.
 * If the function returns -1, no device is found.
 * It returns 1 if a device is found and at least one can be found on the next
 * call.
 * Finaly it return 0 for the last found device.
 * To get the ROM codes of all devices, simply call DS1WSearchRom until it
 * returns 0 or -1.
 */
int DS1WSearchRom(char *ROM)
{
    static u_char lastDis = 0;
    u_char disMarker;
    u_char bits;
    u_char index;

    if (!DS1WInit()) {
        lastDis = 0;
        return -1;
    }

    disMarker = 0;
    ROM--;
    DS1WWriteByte(DS1W_SEARCH_ROM);

    for (index = 1; index <= 64; index++) {     /* 8 bytes rom code */
        if ((index - 1) % 8 == 0)
            ROM++;
        bits = 0;
        if (DS1WReadBit())
            bits |= 0x01;
        if (DS1WReadBit())
            bits |= 0x02;
        if (bits == 0x03) {
            lastDis = 0;
            return 0;
        }

        if (bits == 0) {
            if (lastDis == index) {
                bits = 1;
            } else {
                if (index > lastDis) {
                    bits = 0;
                    disMarker = index;
                } else {
                    bits = *(u_char*)ROM & 0x01;
                    if (!bits)
                        disMarker = index;
                }
            }
        } else
            bits &= 0x01;

        *(u_char*)ROM >>= 1;
        if (bits)
            *(u_char*)ROM |= 0x80;

        DS1WWriteBit(bits);
    }
    lastDis = disMarker;
    if (lastDis) 
        return 1;
    else
        return 0;
}

